import sys
import json
import socketserver

import requests
from cachecontrol import CacheControl
from cachecontrol.heuristics import ExpiresAfter

import ldapserver
from ldapserver import LDAPRequestHandler, rfc4518_stringprep
from ldapserver.dn import DN
from ldapserver.exceptions import LDAPInvalidCredentials
from ldapserver.schema import RFC2307BIS_SUBSCHEMA, RFC2798_SUBSCHEMA, WILDCARD_VALUE

memberOf = ldapserver.schema.AttributeType(
	'1.2.840.113556.1.2.102',
	name='memberOf',
	desc='Group that the entry belongs to',
	equality=ldapserver.schema.rfc4517.matching_rules.distinguishedNameMatch,
	syntax=ldapserver.schema.rfc4517.syntaxes.DN(),
	usage=ldapserver.schema.AttributeTypeUsage.dSAOperation
)

CUSTOM_SUBSCHEMA = RFC2307BIS_SUBSCHEMA.extend(RFC2798_SUBSCHEMA, attribute_types=[memberOf])

class UffdAPI:
	def __init__(self, baseurl, key, cache_ttl=60):
		self.baseurl = baseurl
		self.key = key
		self.session = CacheControl(requests.Session(), heuristic=ExpiresAfter(seconds=cache_ttl))
		self.session.headers['Authorization'] = 'Bearer '+self.key

	def get(self, endpoint, **kwargs):
		resp = self.session.get(self.baseurl + '/' + endpoint, params=kwargs)
		assert(resp.ok)
		return resp.json()

	def post(self, endpoint, **kwargs):
		resp = self.session.post(self.baseurl + '/' + endpoint, data=kwargs)
		assert(resp.ok)
		return resp.json()

	def get_users(self, id=None, loginname=None, group=None):
		return self.get('getusers', id=id, loginname=loginname, group=group)

	def get_groups(self, id=None, name=None, member=None):
		return self.get('getgroups', id=id, name=name, member=member)

	def check_password(self, loginname, password):
		return self.api.post('checkpassword', loginname=loginname, password=password)

def normalize_user_loginname(loginname):
	# The equality matching rule for uid is caseIgnoreMatch. It prepares
	# attribute and assertion value according to LDAP stringprep with
	# case-folding.
	#
	# Uffd restricts loginnames to lower-case ASCII letters, digits,
	# underscores and dashes. None of these characters are changed or
	# rejetced by stringprep with case-folding. The effect stringprep has
	# on loginnames is that it adds a leading and a final SPACE character.
	#
	# The assertion value (the argument to this function) could however contain
	# characters that are mapped to SPACE or nothing for example. So we apply
	# stringprep to the assertion value and then strip the added leading and
	# final SPACE characters. Stringprep case-folds the input string to
	# lower-case.
	#
	# The resulting string can be compared to an actual loginname with simple
	# byte-for-byte or codepoint-for-codepoint comparison with or without
	# case-folding. It matches if and only if the loginname matches the input
	# value according to caseIgnoreMatch.
	try:
		return rfc4518_stringprep.prepare(loginname, rfc4518_stringprep.MatchingType.CASE_IGNORE_STRING).strip(' ')
	except ValueError: # Input value contains prohibited characters
		return None

def normalize_group_name(name):
	# Currently uffd has no restrictions for group names, but it is planned
	# to add restrictions similar to loginname restrictions.
	# See https://git.cccv.de/uffd/uffd/-/issues/127
	return normalize_user_loginname(name)

class RequestHandler(LDAPRequestHandler):
	subschema = CUSTOM_SUBSCHEMA

	# Overwritten before use
	api = None
	dn_base = None
	bind_dn = None
	bind_password = None

	def setup(self):
		super().setup()

	def do_bind_simple_authenticated(self, dn, password):
		dn = DN.from_str(dn)
		if dn == self.bind_dn and password == self.bind_password:
			return True
		if not dn.is_direct_child_of(DN('ou=users') + self.dn_base) or len(dn[0]) != 1 or dn[0][0].attribute != 'uid':
			raise LDAPInvalidCredentials()
		if self.api.check_password(loginname=dn[0][0].value, password=password):
			return True
		raise LDAPInvalidCredentials()

	supports_sasl_plain = True

	def do_bind_sasl_plain(self, identity, password, authzid=None):
		if authzid is not None and identity != authzid:
			raise LDAPInvalidCredentials()
		user = self.api.check_password(loginname=identity, password=password)
		if user is None:
			raise LDAPInvalidCredentials()
		return user

	def do_search(self, baseobj, scope, filter):
		yield from super().do_search(baseobj, scope, filter)
		if self.bind_object:
			yield from self.do_search_static()
			yield from self.do_search_users(baseobj, scope, filter)
			yield from self.do_search_groups(baseobj, scope, filter)

	def do_search_static(self):
		base_attrs = {
			'objectClass': ['top', 'dcObject', 'organization'],
			'structuralObjectClass': ['organization'],
		}
		for rdnassertion in self.dn_base[0]:
			base_attrs[rdnassertion.attribute] = [rdnassertion.value]
		yield self.subschema.Object(self.dn_base, **base_attrs)
		yield self.subschema.Object(DN('ou=users') + self.dn_base,
			ou=['users'],
			objectClass=['top', 'organizationalUnit'],
			structuralObjectClass=['organizationalUnit'],
		)
		yield self.subschema.Object(DN('ou=groups') + self.dn_base,
			ou=['groups'],
			objectClass=['top', 'organizationalUnit'],
			structuralObjectClass=['organizationalUnit'],
		)
		yield self.subschema.Object(DN('ou=system') + self.dn_base,
			ou=['system'],
			objectClass=['top', 'organizationalUnit'],
			structuralObjectClass=['organizationalUnit'],
		)
		yield self.subschema.Object(DN('cn=service,ou=system') + self.dn_base,
			cn=['service'],
			objectClass=['top', 'organizationalRole', 'simpleSecurityObject'],
			structuralObjectClass=['organizationalRole'],
		)

	def do_search_users(self, baseobj, scope, filter):
		template = self.subschema.ObjectTemplate(DN(self.dn_base, ou='users'), 'uid',
			structuralObjectClass=['inetorgperson'],
			objectClass=['top', 'inetorgperson', 'organizationalperson', 'person', 'posixaccount'],
			cn=[WILDCARD_VALUE],
			displayname=[WILDCARD_VALUE],
			givenname=[WILDCARD_VALUE],
			homeDirectory=[WILDCARD_VALUE],
			mail=[WILDCARD_VALUE],
			sn=[' '],
			uid=[WILDCARD_VALUE],
			uidNumber=[WILDCARD_VALUE],
			memberOf=[WILDCARD_VALUE],
		)
		if not template.match_search(baseobj, scope, filter):
			return
		constraints = template.extract_search_constraints(baseobj, scope, filter)
		request_params = {}
		if 'uid' in constraints:
			request_params = {'loginname': normalize_user_loginname(constraints['uid'][0])}
		elif 'uidnumber' in constraints:
			request_params = {'id': constraints['uidnumber'][0]}
		elif 'memberof' in constraints:
			for value in constraints['memberof']:
				if value.is_direct_child_of(DN(self.dn_base, ou='groups')) and value.object_attribute == 'cn':
					request_params = {'group': normalize_group_name(value.object_value)}
					break
		for user in self.api.get_users(**request_params):
			yield template.create_object(user['loginname'],
				cn=[user['displayname']],
				displayname=[user['displayname']],
				givenname=[user['displayname']],
				homeDirectory=['/home/'+user['loginname']],
				mail=[user['email']],
				uid=[user['loginname']],
				uidNumber=[user['id']],
				memberOf=[DN(DN(self.dn_base, ou='groups'), cn=group) for group in user['groups']],
			)

	def do_search_groups(self, baseobj, scope, filter):
		template = self.subschema.ObjectTemplate(DN(self.dn_base, ou='groups'), 'cn',
			structuralObjectClass=['groupOfUniqueNames'],
			objectClass=['top', 'groupOfUniqueNames', 'posixGroup'],
			cn=[WILDCARD_VALUE],
			description=[' '],
			gidNumber=[WILDCARD_VALUE],
			uniqueMember=[WILDCARD_VALUE],
		)
		if not template.match_search(baseobj, scope, filter):
			return
		constraints = template.extract_search_constraints(baseobj, scope, filter)
		request_params = {}
		if 'cn' in constraints:
			request_params = {'name': normalize_group_name(constraints['cn'][0])}
		elif 'gidnumber' in constraints:
			request_params = {'id': constraints['gidnumber'][0]}
		elif 'uniquemember' in constraints:
			for value in constraints['uniquemember']:
				if value.is_direct_child_of(DN(self.dn_base, ou='users')) and value.object_attribute == 'uid':
					request_params = {'member': normalize_user_loginname(value.object_value)}
					break
		for group in self.api.get_groups(**request_params):
			yield template.create_object(group['name'],
				cn=[group['name']],
				gidNumber=[group['id']],
				uniqueMember=[DN(DN(self.dn_base, ou='users'), uid=user) for user in group['members']],
			)

def main(config):
	dn_base = DN.from_str(config['dn_base'])
	api = UffdAPI(config['api_baseurl'], config['api_key'], config.get('cache_ttl', 60))

	subschema = RFC2307BIS_SUBSCHEMA

	class CustomRequestHandler(RequestHandler):
		pass

	CustomRequestHandler.api = api
	CustomRequestHandler.dn_base = dn_base
	CustomRequestHandler.bind_dn = DN('cn=service,ou=system') + dn_base
	CustomRequestHandler.bind_password = config['bind_password'].encode()

	if config['listen_addr'].startswith('unix:'):
		socketserver.ThreadingUnixStreamServer(config['listen_addr'][5:], CustomRequestHandler).serve_forever()
	else:
		addr = config['listen_addr']
		port = '389'
		if addr.startswith('['):
			addr, remainder = addr[1:].split(']', 1)
			if remainder.startswith(':'):
				port = remainder[1:]
		elif ':' in addr:
			addr, port = addr.split(':')
		socketserver.ThreadingTCPServer((addr, int(port)), CustomRequestHandler).serve_forever()

if __name__ == '__main__':
	if len(sys.argv) != 2:
		print('usage: server.py CONFIG_PATH')
		exit(1)
	with open(sys.argv[1], 'r') as f:
		config = json.load(f)
	main(config)
